@@ -16,8 +16,6 @@
 
 import functools
 
-from positional import positional
-
 from keystoneauth1 import _utils as utils
 from keystoneauth1.access import service_catalog
 from keystoneauth1.access import service_providers
@@ -33,7 +31,6 @@ __all__ = ('AccessInfo',
            'create')
 
 
-@positional()
 def create(resp=None, body=None, auth_token=None):
     if resp and not body:
         body = resp.json()
@@ -397,6 +394,14 @@ class AccessInfo(object):
         """
         raise NotImplementedError()
 
+    @property
+    def project_is_domain(self):
+        """Return if a project act as a domain.
+
+        :returns: bool
+        """
+        raise NotImplementedError()
+
 
 class AccessInfoV2(AccessInfo):
     """An object for encapsulating raw v2 auth token from identity service."""
@@ -528,6 +533,10 @@ class AccessInfoV2(AccessInfo):
         except KeyError:
             pass
 
+    @property
+    def project_is_domain(self):
+        return False
+
     @property
     def project_domain_id(self):
         return None
@@ -658,6 +667,10 @@ class AccessInfoV3(AccessInfo):
     def project_id(self):
         return self._project['id']
 
+    @_missingproperty
+    def project_is_domain(self):
+        return self._data['token']['is_domain']
+
     @_missingproperty
     def project_domain_id(self):
         return self._project['domain']['id']
