@@ -24,7 +24,6 @@ raw data specified in version discovery responses.
 import copy
 import re
 
-from positional import positional
 import six
 from six.moves import urllib
 
@@ -57,7 +56,6 @@ def _int_or_latest(val):
     return LATEST if val == 'latest' or val == LATEST else int(val)
 
 
-@positional()
 def get_version_data(session, url, authenticated=None):
     """Retrieve raw version data from a url.
 
@@ -372,8 +370,31 @@ def _combine_relative_url(discovery_url, version_url):
     # effect on absolute urls.
     url = urllib.parse.urljoin(discovery_url.rstrip('/') + '/', version_url)
 
-    # Parse and recombine the result to squish double //'s from the above
-    return urllib.parse.urlparse(url).geturl()
+    # Sadly version discovery documents are common with the scheme
+    # and netloc broken.
+    parsed_version_url = urllib.parse.urlparse(url)
+    parsed_discovery_url = urllib.parse.urlparse(discovery_url)
+
+    return urllib.parse.ParseResult(
+        parsed_discovery_url.scheme,
+        parsed_discovery_url.netloc,
+        parsed_version_url.path,
+        parsed_version_url.params,
+        parsed_version_url.query,
+        parsed_version_url.fragment).geturl()
+
+
+def _version_from_url(url):
+    if not url:
+        return url
+
+    url = urllib.parse.urlparse(url)
+    for part in reversed(url.path.split('/')):
+        try:
+            return normalize_version_number(part)
+        except Exception:
+            pass
+    return None
 
 
 class Discover(object):
@@ -382,7 +403,6 @@ class Discover(object):
     DEPRECATED_STATUSES = ('deprecated',)
     EXPERIMENTAL_STATUSES = ('experimental',)
 
-    @positional()
     def __init__(self, session, url, authenticated=None):
         self._url = url
         self._data = get_version_data(session, url,
@@ -428,7 +448,6 @@ class Discover(object):
 
         return versions
 
-    @positional()
     def version_data(self, reverse=False, **kwargs):
         """Get normalized version data.
 
@@ -660,7 +679,6 @@ class EndpointData(object):
     possibilities.
     """
 
-    @positional()
     def __init__(self,
                  catalog_url=None,
                  service_url=None,
@@ -686,7 +704,6 @@ class EndpointData(object):
         self.region_name = region_name
         self.endpoint_id = endpoint_id
         self.raw_endpoint = raw_endpoint
-        self.api_version = api_version
         self.major_version = major_version
         self.min_microversion = min_microversion
         self.max_microversion = max_microversion
@@ -697,6 +714,8 @@ class EndpointData(object):
         self._catalog_matches_exactly = False
         self._disc = None
 
+        self.api_version = api_version or _version_from_url(self.url)
+
     def __copy__(self):
         """Return a new EndpointData based on this one."""
         new_data = EndpointData(
@@ -725,7 +744,40 @@ class EndpointData(object):
     def url(self):
         return self.service_url or self.catalog_url
 
-    @positional(3)
+    def get_current_versioned_data(self, session, allow=None, cache=None,
+                                   project_id=None):
+        """Run version discovery on the current endpoint.
+
+        A simplified version of get_versioned_data, get_current_versioned_data
+        runs discovery but only on the endpoint that has been found already.
+
+        It can be useful in some workflows where the user wants version
+        information about the endpoint they have.
+
+        :param session: A session object that can be used for communication.
+        :type session: keystoneauth1.session.Session
+        :param dict allow: Extra filters to pass when discovering API
+                           versions. (optional)
+        :param dict cache: A dict to be used for caching results in
+                           addition to caching them on the Session.
+                           (optional)
+        :param string project_id: ID of the currently scoped project. Used for
+                                  removing project_id components of URLs from
+                                  the catalog. (optional)
+
+        :returns: A new EndpointData with the requested versioned data.
+        :rtype: :py:class:`keystoneauth1.discover.EndpointData`
+        :raises keystoneauth1.exceptions.discovery.DiscoveryFailure: If the
+                                                    appropriate versioned data
+                                                    could not be discovered.
+        """
+        min_version, max_version = _normalize_version_args(
+            self.api_version, None, None)
+        return self.get_versioned_data(
+            session=session, allow=allow, cache=cache, allow_version_hack=True,
+            discover_versions=True,
+            min_version=min_version, max_version=max_version)
+
     def get_versioned_data(self, session, allow=None, cache=None,
                            allow_version_hack=True, project_id=None,
                            discover_versions=True,
@@ -843,6 +895,7 @@ class EndpointData(object):
         self.max_microversion = discovered_data['max_microversion']
         self.next_min_version = discovered_data['next_min_version']
         self.not_before = discovered_data['not_before']
+        self.api_version = discovered_data['version']
 
         # TODO(mordred): these next two things should be done by Discover
         # in versioned_data_for.
@@ -863,7 +916,6 @@ class EndpointData(object):
                                        self._saved_project_id)
         self.service_url = url
 
-    @positional(1)
     def _run_discovery(self, session, cache, min_version, max_version,
                        project_id, allow_version_hack, discover_versions):
         tried = set()
@@ -933,7 +985,7 @@ class EndpointData(object):
         min_version and max_version are already normalized, so will either be
         None or a tuple.
         """
-        url = urllib.parse.urlparse(self.url)
+        url = urllib.parse.urlparse(self.url.rstrip('/'))
         url_parts = url.path.split('/')
 
         # First, check to see if the catalog url ends with a project id
@@ -946,7 +998,7 @@ class EndpointData(object):
             try:
                 normalize_version_number(url_parts[-2])
                 self._saved_project_id = url_parts.pop()
-            except TypeError:
+            except (IndexError, TypeError):
                 pass
 
         catalog_discovery = versioned_discovery = None
@@ -988,7 +1040,7 @@ class EndpointData(object):
                     '/'.join(url_parts),
                     url.params,
                     url.query,
-                    url.fragment).geturl()
+                    url.fragment).geturl().rstrip('/') + '/'
 
             # If we found a viable catalog endpoint and it's
             # an exact match or matches the max, go ahead and give
@@ -1002,13 +1054,20 @@ class EndpointData(object):
         if allow_version_hack:
             # If there were projects or versions in the url they are now gone.
             # That means we're left with what should be the unversioned url.
-            yield urllib.parse.ParseResult(
+            hacked_url = urllib.parse.ParseResult(
                 url.scheme,
                 url.netloc,
                 '/'.join(url_parts),
                 url.params,
                 url.query,
                 url.fragment).geturl()
+            # Since this is potentially us constructing a base URL from the
+            # versioned URL - we need to make sure it has a trailing /. But
+            # we only want to do that if we have built a new URL - not if
+            # we're using the one from the catalog
+            if hacked_url != self.catalog_url:
+                hacked_url = hacked_url.strip('/') + '/'
+            yield hacked_url
 
             # If we have a catalog discovery url, it either means we didn't
             # return it earlier because it wasn't an exact enough match, or
@@ -1046,7 +1105,6 @@ class EndpointData(object):
         return _VERSION_HACKS.get_discover_hack(self.service_type, self.url)
 
 
-@positional()
 def get_discovery(session, url, cache=None, authenticated=False):
     """Return the discovery object for a URL.
 
