@@ -22,13 +22,13 @@ import sys
 import time
 import uuid
 
-from positional import positional
 import requests
 import six
 from six.moves import urllib
 
 import keystoneauth1
 from keystoneauth1 import _utils as utils
+from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 
 try:
@@ -65,6 +65,22 @@ def _construct_session(session_obj=None):
     return session_obj
 
 
+def _mv_legacy_headers_for_service(mv_service_type):
+    """Workaround for services that predate standardization.
+
+    TODO(sdague): eventually convert this to using os-service-types
+    and put the logic there. However, right now this is so little
+    logic, inlining it for release is a better call.
+
+    """
+    headers = []
+    if mv_service_type == "compute":
+        headers.append("X-OpenStack-Nova-API-Version")
+    elif mv_service_type == "baremetal":
+        headers.append("X-OpenStack-Ironic-API-Version")
+    return headers
+
+
 class _JSONEncoder(json.JSONEncoder):
 
     def default(self, o):
@@ -221,6 +237,13 @@ class Session(object):
                                        will be added to the user agent. This
                                        can be used by libraries that are part
                                        of the communication process.
+    :param dict discovery_cache: A dict to be used for caching of discovery
+                                 information. This is normally managed
+                                 transparently, but if the user wants to
+                                 share a single cache across multiple sessions
+                                 that do not share an auth plugin, it can
+                                 be provided here. (optional, defaults to
+                                 None which means automatically manage)
     """
 
     user_agent = None
@@ -229,11 +252,11 @@ class Session(object):
 
     _DEFAULT_REDIRECT_LIMIT = 30
 
-    @positional(2)
     def __init__(self, auth=None, session=None, original_ip=None, verify=True,
                  cert=None, timeout=None, user_agent=None,
                  redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
-                 app_name=None, app_version=None, additional_user_agent=None):
+                 app_name=None, app_version=None, additional_user_agent=None,
+                 discovery_cache=None):
 
         self.auth = auth
         self.session = _construct_session(session)
@@ -247,6 +270,9 @@ class Session(object):
         self.app_version = app_version
         self.additional_user_agent = additional_user_agent or []
         self._determined_user_agent = None
+        if discovery_cache is None:
+            discovery_cache = {}
+        self._discovery_cache = discovery_cache
 
         if timeout is not None:
             self.timeout = float(timeout)
@@ -298,7 +324,6 @@ class Session(object):
             return (header[0], '{SHA1}%s' % token_hash)
         return header
 
-    @positional()
     def _http_log_request(self, url, method=None, data=None,
                           json=None, headers=None, query_params=None,
                           logger=_logger):
@@ -346,7 +371,6 @@ class Session(object):
 
         logger.debug(' '.join(string_parts))
 
-    @positional()
     def _http_log_response(self, response=None, json=None,
                            status_code=None, headers=None, text=None,
                            logger=_logger):
@@ -366,8 +390,15 @@ class Session(object):
                 # stream of bytes and getting an unexpected MemoryError. See
                 # bug 1616105 for further details.
                 content_type = response.headers.get('content-type', None)
-                if content_type in _LOG_CONTENT_TYPES:
-                    text = self._remove_service_catalog(response.text)
+
+                # NOTE(lamt): Per [1], the Content-Type header can be of the
+                # form Content-Type := type "/" subtype *[";" parameter]
+                # [1] https://www.w3.org/Protocols/rfc1341/4_Content-Type.html
+                for log_type in _LOG_CONTENT_TYPES:
+                    if content_type is not None and content_type.startswith(
+                            log_type):
+                        text = self._remove_service_catalog(response.text)
+                        break
                 else:
                     text = ('Omitted, Content-Type is set to %s. Only '
                             '%s responses have their bodies logged.')
@@ -387,13 +418,57 @@ class Session(object):
 
         logger.debug(' '.join(string_parts))
 
-    @positional()
+    @staticmethod
+    def _set_microversion_headers(
+            headers, microversion, service_type, endpoint_filter):
+        # We're converting it to normalized version number for two reasons.
+        # First, to validate it's a real version number. Second, so that in
+        # the future we can pre-validate that it is within the range of
+        # available microversions before we send the request.
+        # TODO(mordred) Validate when we get the response back that
+        # the server executed in the microversion we expected.
+        # TODO(mordred) Validate that the requested microversion works
+        # with the microversion range we found in discovery.
+        microversion = discover.normalize_version_number(microversion)
+        # Can't specify a M.latest microversion
+        if (microversion[0] != discover.LATEST and
+                discover.LATEST in microversion[1:]):
+            raise TypeError(
+                "Specifying a '{major}.latest' microversion is not allowed.")
+        microversion = discover.version_to_string(microversion)
+        if not service_type:
+            if endpoint_filter and 'service_type' in endpoint_filter:
+                service_type = endpoint_filter['service_type']
+            else:
+                raise TypeError(
+                    "microversion {microversion} was requested but no"
+                    " service_type information is available. Either provide a"
+                    " service_type in endpoint_filter or pass"
+                    " microversion_service_type as an argument.".format(
+                        microversion=microversion))
+
+        # TODO(mordred) cinder uses volume in its microversion header. This
+        # logic should be handled in the future by os-service-types but for
+        # now hard-code for cinder.
+        if (service_type.startswith('volume')
+                or service_type == 'block-storage'):
+            service_type = 'volume'
+        headers.setdefault('OpenStack-API-Version',
+                           '{service_type} {microversion}'.format(
+                               service_type=service_type,
+                               microversion=microversion))
+        header_names = _mv_legacy_headers_for_service(service_type)
+        for h in header_names:
+            headers.setdefault(h, microversion)
+
     def request(self, url, method, json=None, original_ip=None,
                 user_agent=None, redirect=None, authenticated=None,
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
                 endpoint_override=None, connect_retries=0, logger=_logger,
-                allow={}, client_name=None, client_version=None, **kwargs):
+                allow=None, client_name=None, client_version=None,
+                microversion=None, microversion_service_type=None,
+                **kwargs):
         """Send an HTTP request with the specified characteristics.
 
         Wrapper around `requests.Session.request` to handle tasks such as
@@ -461,6 +536,16 @@ class Session(object):
         :type logger: logging.Logger
         :param dict allow: Extra filters to pass when discovering API
                            versions. (optional)
+        :param microversion: Microversion to send for this request.
+                       microversion can be given as a string or a tuple.
+                       (optional)
+        :param str microversion_service_type: The service_type to be sent in
+                       the microversion header, if a microversion is given.
+                       Defaults to the value of service_type from
+                       endpoint_filter if one exists. If endpoint_filter is not
+                       provided or does not have a service_type, microversion
+                       is given and microversion_service_type is not provided,
+                       an exception will be raised.
         :param kwargs: any other parameter that can be passed to
                        :meth:`requests.Session.request` (such as `headers`).
                        Except:
@@ -476,6 +561,10 @@ class Session(object):
         :returns: The response to the request.
         """
         headers = kwargs.setdefault('headers', dict())
+        if microversion:
+            self._set_microversion_headers(
+                headers, microversion, microversion_service_type,
+                endpoint_filter)
 
         if authenticated is None:
             authenticated = bool(auth or self.auth)
@@ -608,6 +697,30 @@ class Session(object):
 
         resp = send(**kwargs)
 
+        # log callee and caller request-id for each api call
+        if log:
+            # service_name should be fetched from endpoint_filter if it is not
+            # present then use service_type as service_name.
+            service_name = None
+            if endpoint_filter:
+                service_name = endpoint_filter.get('service_name')
+                if not service_name:
+                    service_name = endpoint_filter.get('service_type')
+
+            # Nova uses 'x-compute-request-id' and other services like
+            # Glance, Cinder etc are using 'x-openstack-request-id' to store
+            # request-id in the header
+            request_id = (resp.headers.get('x-openstack-request-id') or
+                          resp.headers.get('x-compute-request-id'))
+            if request_id:
+                logger.debug('%(method)s call to %(service_name)s for '
+                             '%(url)s used request id '
+                             '%(response_request_id)s',
+                             {'method': resp.request.method,
+                              'service_name': service_name,
+                              'url': resp.url,
+                              'response_request_id': request_id})
+
         # handle getting a 401 Unauthorized response by invalidating the plugin
         # and then retrying the request. This is only tried once.
         if resp.status_code == 401 and authenticated and allow_reauth:
@@ -821,9 +934,46 @@ class Session(object):
         :returns: An endpoint if available or None.
         :rtype: string
         """
+        if 'endpoint_override' in kwargs:
+            return kwargs['endpoint_override']
+
         auth = self._auth_required(auth, 'determine endpoint URL')
+
         return auth.get_endpoint(self, **kwargs)
 
+    def get_endpoint_data(self, auth=None, **kwargs):
+        """Get endpoint data as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
+        :raises TypeError: If arguments are invalid
+
+        :returns: Endpoint data if available or None.
+        :rtype: keystoneauth1.discover.EndpointData
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_endpoint_data(self, **kwargs)
+
+    def get_api_major_version(self, auth=None, **kwargs):
+        """Get the major API version as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
+            plugin is not available.
+
+        :return: The major version of the API of the service discovered.
+        :rtype: tuple or None
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_api_major_version(self, **kwargs)
+
     def get_auth_connection_params(self, auth=None, **kwargs):
         """Return auth connection params as provided by the auth plugin.
 
@@ -933,7 +1083,7 @@ class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
 
     This Adapter also preserves the default behaviour of Requests which
     disables Nagle's Algorithm. See also:
-    http://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
+    https://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
     """
 
     def init_poolmanager(self, *args, **kwargs):
